// zeroVM - VM crash course
// How to build a simple virtual machine
// Stefan Pietzonke 2019
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

typedef unsigned char           U1;		/* UBYTE   */
typedef int16_t                 S2;     /* INT     */
typedef uint16_t                U2;     /* UINT    */
typedef int32_t                 S4;     /* LONGINT */

typedef long long               S8;     /* 64 bit long */
typedef double                  F8;     /* DOUBLE */


#define DEBUG 1

#define EXE_NEXT(); ep = ep + eoffs; goto *jumpt[code[ep]];

/* opcodes */
#define PUSHB	0
#define PULLB	1

#define LOADB 	2
#define ADDI	3

#define PRINTI	4

#define PEXIT	5

//	space for code and data
unsigned char code[1024];
unsigned char data[1024];

int main (int ac, char *av[])
{
	S8 ep;				// execution pointer
	S8 eoffs = 0;		// number of bytes to skip to get next opcode

	S8 regi[256];			// integer registers 64 bit
	S8 arg1, arg2, arg3;	// opcode arguments

	S8 cpu_core = 0;

	printf ("zeroVM\n");

	// data bytes 23 and 42
	data[0] = 23;
	data[1] = 42;

	// code
	// load byte from data_size
	code[0] = LOADB;
	code[1] = 0;	// data ptr
	code[2] = 0;	// offset
	code[3] = 0;	// target register

	// load byte from data
	code[4] = LOADB;
	code[5] = 1;
	code[6] = 0;
	code[7] = 1;

	code[8] = ADDI;
	code[9] = 0;	// register 0: 23
	code[10] = 1;	// register 1: 42
	code[11] = 2;	// target register 2: = 65

	code[12] = PRINTI;
	code[13] = 2;	// register 2: 65

	code[14] = PEXIT;

	// execution bytecode pointer
	ep = 0;

	// jumptable for indirect threading execution
	static void *jumpt[] =
    {
		&&pushb, &&pullb, &&loadb, &&addi, &&printi, &&pexit
	};

	printf ("running test program...\n\n");

	// fetch first bytecode
	EXE_NEXT();


	pushb:
	#if DEBUG
	printf ("%lli PUSHB\n", cpu_core);
	#endif
	arg1 = regi[code[ep + 1]];
	arg2 = regi[code[ep + 2]];
	arg3 = code[ep + 3];

	regi[arg3] = data[arg1 + arg2];

	eoffs = 4;
	EXE_NEXT();


	pullb:
	#if DEBUG
	printf ("%lli PULLB\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = regi[code[ep + 2]];
	arg3 = regi[code[ep + 3]];

	data[arg2 + arg3] = regi[arg1];

	eoffs = 4;
	EXE_NEXT();


	loadb:
	#if DEBUG
	printf ("%lli LOADB\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = data[arg1 + arg2];

	eoffs = 4;
	EXE_NEXT();


	addi:
	#if DEBUG
	printf ("%lli ADDI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] + regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	printi:
	#if DEBUG
	printf ("PRINTI\n");
	#endif
	arg1 = code[ep + 1];

	printf ("%lli\n", regi[arg1]);

	eoffs = 2;
	EXE_NEXT();

	pexit:
	printf ("shutdown...\n");
	exit (0);
}
